package net.knarcraft.placeholdersigns.listener;

import net.knarcraft.knarlib.formatting.StringFormatter;
import net.knarcraft.knarlib.property.ColorConversion;
import net.knarcraft.knarlib.util.ColorHelper;
import net.knarcraft.placeholdersigns.PlaceholderSigns;
import net.knarcraft.placeholdersigns.config.PlaceholderSignMessage;
import net.knarcraft.placeholdersigns.container.PlaceholderSign;
import net.knarcraft.placeholdersigns.container.SignCopyRequest;
import net.knarcraft.placeholdersigns.container.SignLineChangeRequest;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignHandler;
import net.knarcraft.placeholdersigns.handler.PlaceholderSignRequestHandler;
import org.bukkit.Bukkit;
import org.bukkit.block.Sign;
import org.bukkit.block.sign.Side;
import org.bukkit.block.sign.SignSide;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.block.SignChangeEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Map;
import java.util.Objects;

/**
 * A listener for placeholder signs being clicked
 */
public class SignClickListener implements Listener {

    @EventHandler(priority = EventPriority.MONITOR)
    public void onSignClick(@NotNull PlayerInteractEvent event) {
        // Ignore if not a clicked sign
        if (!event.hasBlock() || event.getClickedBlock() == null ||
                !(event.getClickedBlock().getState() instanceof Sign sign)) {
            return;
        }

        Player player = event.getPlayer();
        PlaceholderSignRequestHandler requestHandler = PlaceholderSigns.getInstance().getRequestHandler();

        SignCopyRequest signCopyRequest = requestHandler.getSignCopyRequest(player);
        if (signCopyRequest != null) {
            if (checkWaxEdit(sign, player, "placeholdersigns.copy.bypass-waxed")) {
                return;
            }

            pasteSign(signCopyRequest.sign(), sign, player);
            event.setCancelled(true);
            return;
        }

        // Check if the player has run the /editSign command
        SignLineChangeRequest request = requestHandler.getSignChangeRequest(player);
        if (request != null) {
            if (checkWaxEdit(sign, player, "placeholdersigns.edit.bypass-waxed")) {
                return;
            }

            SignSide standingOn = sign.getTargetSide(player);
            Side side = sign.getSide(Side.FRONT).equals(standingOn) ? Side.FRONT : Side.BACK;
            doSignChange(sign, request, event, player, side);
        }
    }

    /**
     * Checks if the player is trying to edit/modify a waxed sign without permission
     *
     * @param sign   <p>The sign the player is trying to edit</p>
     * @param player <p>The player trying to edit the sign</p>
     * @return <p>True if the player is trying to edit a waxed sign without permission</p>
     */
    private boolean checkWaxEdit(@NotNull Sign sign, @NotNull Player player, @NotNull String permissionNode) {
        if (sign.isWaxed() && !player.hasPermission(permissionNode)) {
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_WAXED_NO_PERMISSION);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Pastes the data from the source sign to the target sign
     *
     * @param source <p>The source sign to copy data from</p>
     * @param target <p>The target sign to paste to</p>
     * @param player <p>The player pasting the sign</p>
     */
    private void pasteSign(@NotNull Sign source, @NotNull Sign target, @NotNull Player player) {
        StringFormatter stringFormatter = PlaceholderSigns.getInstance().getStringFormatter();
        PlaceholderSign placeholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(source.getLocation());
        SignSide sourceFront = source.getSide(Side.FRONT);
        SignSide sourceBack = source.getSide(Side.BACK);
        SignSide targetFront = target.getSide(Side.FRONT);
        SignSide targetBack = target.getSide(Side.BACK);

        // Get the lines, with placeholders replaced, and the sign change event processed
        String[] frontLines = getFinalLines(target, sourceFront, Side.FRONT, placeholderSign, player);
        if (frontLines == null) {
            stringFormatter.displayErrorMessage(player, PlaceholderSignMessage.ERROR_CANCELLED_BY_PROTECTION);
            return;
        }
        String[] backLines = getFinalLines(target, sourceBack, Side.BACK, placeholderSign, player);
        if (backLines == null) {
            stringFormatter.displayErrorMessage(player, PlaceholderSignMessage.ERROR_CANCELLED_BY_PROTECTION);
            return;
        }

        // Update the lines on both sides of the target sign
        for (int i = 0; i < frontLines.length; i++) {
            targetFront.setLine(i, ColorHelper.translateColorCodes(frontLines[i], ColorConversion.RGB));
        }
        for (int i = 0; i < backLines.length; i++) {
            targetBack.setLine(i, ColorHelper.translateColorCodes(backLines[i], ColorConversion.RGB));
        }

        // Copy options
        copySignOptions(source, target);

        // Apply changes
        target.update();

        stringFormatter.displaySuccessMessage(player, PlaceholderSignMessage.SUCCESS_SIGN_PASTED);
    }

    /**
     * Gets the final lines from a sign side, after inserting placeholders and running a sign change event
     *
     * @param sign <p>The sign that's changed</p>
     * @param signSide <p>The side of the sign to get lines from</p>
     * @param side <p>The side that's processed</p>
     * @param placeholderSign <p>The placeholder sign corresponding to the sign, if any</p>
     * @param player <p>The player attempting to paste the sign</p>
     * @return <p>The final lines, or null if the event was cancelled</p>
     */
    @Nullable
    private String[] getFinalLines(@NotNull Sign sign, @NotNull SignSide signSide, @NotNull Side side,
                                   @Nullable PlaceholderSign placeholderSign, @NotNull Player player) {
        String[] sideLines = signSide.getLines();
        if (placeholderSign != null) {
            Map<Side, Map<Integer, String>> placeholders = placeholderSign.placeholders();
            loadPlaceholders(side, sideLines, placeholders);
        }

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(sign.getBlock()), player, sideLines, side);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            return null;
        }

        return changeEvent.getLines();
    }

    /**
     * Copies all sign option (dye, glowing, waxed) from the source to the target
     *
     * @param source <p>The source sign</p>
     * @param target <p>The target sign</p>
     */
    private void copySignOptions(@NotNull Sign source, @NotNull Sign target) {
        SignSide sourceFront = source.getSide(Side.FRONT);
        SignSide sourceBack = source.getSide(Side.BACK);
        SignSide targetFront = target.getSide(Side.FRONT);
        SignSide targetBack = target.getSide(Side.BACK);

        // Copy glowing state
        targetFront.setGlowingText(sourceFront.isGlowingText());
        targetBack.setGlowingText(sourceBack.isGlowingText());

        // Copy applied dye
        targetFront.setColor(sourceFront.getColor());
        targetBack.setColor(sourceBack.getColor());

        // Apply waxed state
        target.setWaxed(source.isWaxed());
    }

    /**
     * Loads placeholders from one side of a sign
     *
     * @param side         <p>The side to load from</p>
     * @param lines        <p>The array to write placeholders to</p>
     * @param placeholders <p>The placeholders to load</p>
     */
    private void loadPlaceholders(@NotNull Side side, @NotNull String[] lines,
                                  @NotNull Map<Side, Map<Integer, String>> placeholders) {
        if (!placeholders.containsKey(side)) {
            return;
        }

        Map<Integer, String> sidePlaceholders = placeholders.get(side);
        for (int i = 0; i < 4; i++) {
            if (sidePlaceholders.containsKey(i)) {
                lines[i] = sidePlaceholders.get(i);
            }
        }
    }

    /**
     * Perform the changing of a sign according to the given sign change request
     *
     * @param sign    <p>The sign to be changed</p>
     * @param request <p>The sign line change request to perform</p>
     * @param event   <p>The interaction event that triggered this</p>
     * @param player  <p>The player triggering the sign change</p>
     */
    private void doSignChange(@NotNull Sign sign, @NotNull SignLineChangeRequest request,
                              @NotNull PlayerInteractEvent event, @NotNull Player player, @NotNull Side side) {
        // Cancel the event to prevent vanilla behavior
        event.setCancelled(true);

        SignSide signSide = sign.getSide(side);
        String[] lines = signSide.getLines();

        PlaceholderSignHandler signHandler = PlaceholderSigns.getInstance().getSignHandler();
        PlaceholderSign placeholderSign = PlaceholderSigns.getInstance().getSignHandler().getFromLocation(sign.getLocation());
        String oldPlaceholder = null;
        if (placeholderSign != null) {
            // Remove the old placeholder
            oldPlaceholder = placeholderSign.placeholders().get(side).remove(request.line());
        }

        lines[request.line()] = request.text();

        // Run the sign change event to allow protection plugins to cancel, and allow the sign text listener to trigger
        SignChangeEvent changeEvent = new SignChangeEvent(Objects.requireNonNull(event.getClickedBlock()),
                player, lines, side);
        Bukkit.getPluginManager().callEvent(changeEvent);
        if (changeEvent.isCancelled()) {
            if (placeholderSign != null) {
                // Restore the old placeholder if the action didn't complete
                placeholderSign.placeholders().get(side).put(request.line(), oldPlaceholder);
            }
            PlaceholderSigns.getInstance().getStringFormatter().displayErrorMessage(player,
                    PlaceholderSignMessage.ERROR_CANCELLED_BY_PROTECTION);
            return;
        }

        // Update the sign with the new text
        String[] finalLines = changeEvent.getLines();
        for (int i = 0; i < finalLines.length; i++) {
            signSide.setLine(i, ColorHelper.translateColorCodes(finalLines[i], ColorConversion.RGB));
        }
        sign.update();

        // Save any placeholder changes
        signHandler.save();

        PlaceholderSigns.getInstance().getStringFormatter().displaySuccessMessage(player,
                PlaceholderSignMessage.SUCCESS_SIGN_CHANGED);
    }

}
